@@ -7,6 +7,7 @@
 #include "../../cntr.h"
 #include "../../conf.h"
 #include "../../conffile.h"
+#include "../../cstat.h"
 #include "../../fsops.h"
 #include "../../handy.h"
 #include "../../iobuf.h"
@@ -18,6 +19,7 @@
 #include "blocklen.h"
 #include "dpth.h"
 #include "backup_phase2.h"
+#include "link.h"
 
 static size_t treepathlen=0;
 
@@ -30,7 +32,10 @@ static int path_length_warn(struct iobuf *path, struct conf **cconfs)
 
 static int path_too_long(struct iobuf *path, struct conf **cconfs)
 {
-	static const char *cp;
+	const char *cp;
+	const char *cp1;
+	size_t len;
+
 	if(treepathlen+path->len+1>fs_full_path_max)
 	{
 		// FIX THIS:
@@ -41,9 +46,21 @@ static int path_too_long(struct iobuf *path, struct conf **cconfs)
 		// to be able to fix it.
 		return path_length_warn(path, cconfs);
 	}
-	if((cp=strrchr(path->buf, '/'))) cp++;
-	else cp=path->buf;
-	if(strlen(cp)>fs_name_max) return path_length_warn(path, cconfs);
+	// We have to check every part in the path to ensure it less then fs_name_max
+	// minimum windows case is c:/a, nix case is /
+	// usual: c:/users, /home
+	cp = strchr(path->buf, '/');
+	if( !cp ) // very strange
+	    return strlen(path->buf) > fs_name_max ?  path_length_warn(path, cconfs):0;
+	while(cp && *cp)
+	{
+	    cp++;
+	    cp1 = strchr(cp, '/');
+	    len = cp1? cp1-cp : strlen(cp);
+	    if( len > fs_name_max )
+		return path_length_warn(path, cconfs);
+	    cp = cp1;
+	}
 	return 0;
 }
 
@@ -61,8 +78,7 @@ static int treedata(struct sbuf *sb, struct conf **cconfs)
 	return !path_too_long(&sb->path, cconfs);
 }
 
-static char *set_new_datapth(struct asfd *asfd,
-	struct sdirs *sdirs, struct conf **cconfs,
+static char *set_new_datapth(struct sdirs *sdirs, struct conf **cconfs,
 	struct sbuf *sb, struct dpth *dpth, int *istreedata)
 {
 	char *tmp=NULL;
@@ -76,7 +92,7 @@ static char *set_new_datapth(struct asfd *asfd,
 		// the directory structure on the original client.
 		if(!(tmp=prepend_s(TREE_DIR, sb->path.buf)))
 		{
-			log_and_send_oom(asfd);
+			log_out_of_memory(__func__);
 			return NULL;
 		}
 	}
@@ -89,7 +105,7 @@ static char *set_new_datapth(struct asfd *asfd,
 	if(build_path(sdirs->datadirtmp,
 		sb->protocol1->datapth.buf, &rpath, sdirs->datadirtmp))
 	{
-		log_and_send(asfd, "build path failed");
+		logp("build path failed");
 		return NULL;
 	}
 	return rpath;
@@ -103,12 +119,11 @@ static int start_to_receive_new_file(struct asfd *asfd,
 	char *rpath=NULL;
 	int istreedata=0;
 
-//logp("start to receive: %s\n", sb->path.buf);
+//logp("start to receive: %s\n", iobuf_to_printable(&sb->path));
+
+	if(!(rpath=set_new_datapth(sdirs, cconfs, sb, dpth, &istreedata)))
+		goto end;
 
-	if(!(rpath=set_new_datapth(asfd,
-		sdirs, cconfs, sb, dpth, &istreedata)))
-			goto end;
-	
 	if(!(sb->protocol1->fzp=fzp_open(rpath, "wb")))
 	{
 		log_and_send(asfd, "make file failed");
@@ -121,13 +136,25 @@ end:
 	return ret;
 }
 
+enum processed_e
+{
+	P_ERROR=-1,
+	P_DONE_NEW=0,
+	P_DONE_CHANGED=1,
+	P_DONE_UNCHANGED=2,
+	P_DONE_DELETED=3,
+	P_NEW=4,
+	P_CHANGED=5
+};
+
 #include <librsync.h>
 
-static int process_changed_file(struct asfd *asfd, struct conf **cconfs,
+static enum processed_e process_changed_file(struct asfd *asfd,
+	struct conf **cconfs,
 	struct sbuf *cb, struct sbuf *p1b,
 	const char *adir)
 {
-	int ret=-1;
+	int ret=P_ERROR;
 	size_t blocklen=0;
 	char *curpath=NULL;
 	//logp("need to process changed file: %s (%s)\n",
@@ -166,7 +193,7 @@ static int process_changed_file(struct asfd *asfd, struct conf **cconfs,
 		logp("could not start signature job.\n");
 		goto end;
 	}
-	//logp("sig begin: %s\n", p1b->protocol1->datapth.buf);
+	//logp("sig begin: %s\n", iobuf_to_printable(7p1b->protocol1->datapth));
 	if(!(p1b->protocol1->infb=rs_filebuf_new(NULL,
 		p1b->protocol1->sigfzp,
 		NULL, blocklen, -1)))
@@ -189,13 +216,13 @@ static int process_changed_file(struct asfd *asfd, struct conf **cconfs,
 	//logp("sending sig for %s\n", p1b->path);
 	//logp("(%s)\n", p1b->datapth);
 
-	ret=0;
+	ret=P_CHANGED;
 end:
 	free_w(&curpath);
 	return ret;
 }
 
-static int new_non_file(struct sbuf *p1b,
+static enum processed_e new_non_file(struct sbuf *p1b,
 	struct manio *ucmanio, struct conf **cconfs)
 {
 	// Is something that does not need more data backed up.
@@ -204,27 +231,24 @@ static int new_non_file(struct sbuf *p1b,
 	// order with normal files, which has to wait around for their data
 	// to turn up.
 	if(manio_write_sbuf(ucmanio, p1b))
-		return -1;
+		return P_ERROR;
 	cntr_add(get_cntr(cconfs), p1b->path.cmd, 0);
-	sbuf_free_content(p1b);
-	return 0;
+	return P_DONE_NEW;
 }
 
-static int changed_non_file(struct sbuf *p1b,
+static enum processed_e changed_non_file(struct sbuf *p1b,
 	struct manio *ucmanio, enum cmd cmd, struct conf **cconfs)
 {
 	// As new_non_file.
 	if(manio_write_sbuf(ucmanio, p1b))
-		return -1;
+		return P_ERROR;
 	cntr_add_changed(get_cntr(cconfs), cmd);
-	sbuf_free_content(p1b);
-	return 0;
+	return P_DONE_CHANGED;
 }
 
-static int process_new(struct conf **cconfs,
+static enum processed_e process_new(struct conf **cconfs,
 	struct sbuf *p1b, struct manio *ucmanio)
 {
-	if(!p1b->path.buf) return 0;
 	if(sbuf_is_filedata(p1b)
 	  || sbuf_is_vssdata(p1b))
 	{
@@ -232,47 +256,192 @@ static int process_new(struct conf **cconfs,
 		// Flag the things that need to be sent (to the client)
 		p1b->flags |= SBUF_SEND_STAT;
 		p1b->flags |= SBUF_SEND_PATH;
-		return 0;
+		return P_NEW;
 	}
 	return new_non_file(p1b, ucmanio, cconfs);
 }
 
-static int process_unchanged_file(struct sbuf *p1b, struct sbuf *cb,
+static enum processed_e process_unchanged_file(struct sbuf *p1b, struct sbuf *cb,
 	struct manio *ucmanio, struct conf **cconfs)
 {
 	// Need to re-encode the p1b attribs to include compression and
 	// other bits and pieces that are recorded on cb.
 	iobuf_move(&p1b->protocol1->datapth, &cb->protocol1->datapth);
 	iobuf_move(&p1b->endfile, &cb->endfile);
+	p1b->protocol1->salt=cb->protocol1->salt;
 	p1b->compression=cb->compression;
+	p1b->encryption=cb->encryption;
 	// Why is winattr not getting sent in phase1?
 	p1b->winattr=cb->winattr;
 	// Need to free attr so that it is reallocated, because it may get
 	// longer than what the client told us in phase1.
 	iobuf_free_content(&p1b->attr);
 	if(attribs_encode(p1b))
-		return -1;
+		return P_ERROR;
 	if(manio_write_sbuf(ucmanio, p1b))
-		return -1;
+		return P_ERROR;
 	cntr_add_same(get_cntr(cconfs), p1b->path.cmd);
 	if(p1b->endfile.buf) cntr_add_bytes(get_cntr(cconfs),
-		 strtoull(p1b->endfile.buf, NULL, 10));
-	sbuf_free_content(cb);
-	return 1;
+		strtoull(p1b->endfile.buf, NULL, 10));
+	return P_DONE_UNCHANGED;
 }
 
-static int process_new_file(struct conf **cconfs,
-	struct sbuf *cb, struct sbuf *p1b, struct manio *ucmanio)
+static int open_previous_manifest(struct manio **manio, struct sdirs *sdirs)
 {
-	if(process_new(cconfs, p1b, ucmanio))
+	struct stat statp;
+	if(!lstat(sdirs->cmanifest, &statp)
+	  && !(*manio=manio_open(sdirs->cmanifest, "rb", PROTO_1)))
+	{
+		logp("could not open old manifest %s\n", sdirs->cmanifest);
 		return -1;
-	sbuf_free_content(cb);
-	return 1;
+	}
+	return 0;
 }
 
-static int maybe_do_delta_stuff(struct asfd *asfd,
+static int get_hardlink_master_from_hmanio(
+	struct manio **hmanio,
+	struct sdirs *sdirs,
+	struct sbuf *cb,
+	struct sbuf *hb
+) {
+	int passes=0;
+	while(1)
+	{
+		sbuf_free_content(hb);
+		switch(manio_read(*hmanio, hb))
+		{
+			case 0: // Keep going.
+				break;
+			case 1: // Finished OK.
+				manio_close(hmanio);
+				// Might have already been part way down the
+				// manifest when we started. If we get to the
+				// end, go back to the beginning once.
+				if(!passes)
+				{
+					if(open_previous_manifest(hmanio,
+						sdirs)) return -1;
+					passes++;
+					continue;
+				}
+				return 0;
+			default: // Error;
+				return -1;
+		}
+		if(hb->path.cmd!=CMD_FILE
+		  && hb->path.cmd!=CMD_ENC_FILE)
+			continue;
+		if(!strcmp(hb->path.buf, cb->link.buf)
+		  && hb->protocol1->datapth.buf)
+			return 1; // Found it.
+	}
+	return -1;
+}
+
+static int relink_deleted_hardlink_master(
+	struct dpth *dpth,
+	struct manio **hmanio,
+	struct sdirs *sdirs,
+	struct sbuf *p1b,
+	struct sbuf *cb,
+	struct conf **cconfs
+) {
+	int ret=-1;
+	char *newdatapth_full=NULL;
+	char *olddatapth_full=NULL;
+	char *newdatapth=NULL;
+	char *relinkpath_full=NULL;
+	struct stat statp;
+	struct sbuf *hb=NULL;
+	int istreedata=0;
+
+	if(!(hb=sbuf_alloc(PROTO_1)))
+		goto end;
+
+	switch(get_hardlink_master_from_hmanio(hmanio, sdirs, cb, hb))
+	{
+		case 0: // Did not find it.
+			return 0;
+		case 1: // Found it.
+			break;
+		default: // Error.
+			goto end;
+	}
+
+	// Protect against the old file being encrypted, and the new file not
+	// being encrypted - and vice versa.
+	if(p1b->path.cmd!=hb->path.cmd)
+		return 0;
+
+	if(!(newdatapth_full=set_new_datapth(sdirs,
+		cconfs, p1b, dpth, &istreedata)))
+			goto end;
+	if(!istreedata) dpth_incr(dpth);
+	if(!(relinkpath_full=prepend_s(sdirs->relink,
+		p1b->protocol1->datapth.buf)))
+			goto end;
+	if(!(newdatapth=strdup_w(p1b->protocol1->datapth.buf, __func__)))
+		goto end;
+/*
+printf("newdatapth_full: %s\n", newdatapth_full);
+printf("relinkpath_full: %s\n", relinkpath_full);
+printf("newdatapth: %s\n", newdatapth);
+printf("hbdatapth: %s\n", iobuf_to_printable(&hb->protocol1->datapth));
+printf("sdirs->currentdata: %s\n", sdirs->currentdata);
+*/
+
+	if(!(olddatapth_full=prepend_s(sdirs->currentdata,
+		hb->protocol1->datapth.buf)))
+			goto end;
+	if(lstat(olddatapth_full, &statp) || !S_ISREG(statp.st_mode))
+	{
+		logw(NULL, get_cntr(cconfs),
+			"Did not find path for relink: %s\n", olddatapth_full);
+		ret=0;
+		goto end;
+	}
+	if(build_path_w(relinkpath_full))
+		goto end;
+	if(do_link(olddatapth_full, relinkpath_full,
+		&statp, cconfs, /*overwrite*/0))
+			goto end;
+
+	iobuf_free_content(&cb->protocol1->datapth);
+
+	iobuf_from_str(&hb->protocol1->datapth, CMD_DATAPTH,
+		p1b->protocol1->datapth.buf);
+	p1b->protocol1->datapth.buf=NULL;
+	iobuf_from_str(&cb->protocol1->datapth, CMD_DATAPTH, newdatapth);
+	newdatapth=NULL;
+	iobuf_move(&cb->endfile, &hb->endfile);
+
+	ret=1;
+end:
+	free_w(&newdatapth_full);
+	free_w(&olddatapth_full);
+	free_w(&newdatapth);
+	free_w(&relinkpath_full);
+	sbuf_free(&hb);
+	return ret;
+}
+
+static int librsync_enabled(struct sbuf *p1b,
+	struct sbuf *cb,
+	struct conf **cconfs)
+{
+	off_t max_size;
+	if(!get_int(cconfs[OPT_LIBRSYNC]))
+		return 0;
+	max_size=(off_t)get_uint64_t(cconfs[OPT_LIBRSYNC_MAX_SIZE]);
+	if(!max_size)
+		return 1;
+	return max_size >= cb->statp.st_size && max_size >= p1b->statp.st_size;
+}
+
+static enum processed_e maybe_do_delta_stuff(struct asfd *asfd,
+	struct dpth *dpth,
 	struct sdirs *sdirs, struct sbuf *cb, struct sbuf *p1b,
-	struct manio *ucmanio, struct conf **cconfs)
+	struct manio *ucmanio, struct manio **hmanio, struct conf **cconfs)
 {
 	int oldcompressed=0;
 	int compression=p1b->compression;
@@ -280,7 +449,38 @@ static int maybe_do_delta_stuff(struct asfd *asfd,
 	// If the file type changed, I think it is time to back it up again
 	// (for example, EFS changing to normal file, or back again).
 	if(cb->path.cmd!=p1b->path.cmd)
-		return process_new_file(cconfs, cb, p1b, ucmanio);
+	{
+		if(hmanio && *hmanio
+		  && cb->path.cmd==CMD_HARD_LINK
+		  && (p1b->path.cmd==CMD_FILE
+			|| p1b->path.cmd==CMD_ENC_FILE))
+		{
+			struct stat *p1statp=&p1b->statp;
+			struct stat *cstatp=&cb->statp;
+			// A hardlink changed into a file. It is possible that
+			// The file that the hardlink was pointing to got
+			// deleted. Maybe we can reuse the previous file.
+			if(p1statp->st_dev==cstatp->st_dev
+			  && p1statp->st_ino==cstatp->st_ino
+			  && p1statp->st_mtime==cstatp->st_mtime)
+			{
+				switch(relink_deleted_hardlink_master(
+					dpth, hmanio, sdirs,
+					p1b, cb, cconfs))
+				{
+					case 0:
+						break;
+					case 1:
+						return process_unchanged_file(
+							p1b, cb,
+							ucmanio, cconfs);
+					default:
+						return P_ERROR;
+				}
+			}
+		}
+		return process_new(cconfs, p1b, ucmanio);
+	}
 
 	// mtime is the actual file data.
 	// ctime is the attributes or meta data.
@@ -288,8 +488,8 @@ static int maybe_do_delta_stuff(struct asfd *asfd,
 	  && cb->statp.st_ctime==p1b->statp.st_ctime)
 	{
 		// got an unchanged file
-		//logp("got unchanged file: %s %c %c\n",
-		//	cb->path.buf, cb->path.cmd, p1b->path.cmd);
+		//logp("got unchanged file: %s %c\n",
+		//	iobuf_to_printable(&cb->path), p1b->path.cmd);
 		return process_unchanged_file(p1b, cb, ucmanio, cconfs);
 	}
 
@@ -309,7 +509,7 @@ static int maybe_do_delta_stuff(struct asfd *asfd,
 		  || p1b->path.cmd==CMD_METADATA
 		  || sbuf_is_vssdata(cb)
 		  || sbuf_is_vssdata(p1b))
-			return process_new_file(cconfs, cb, p1b, ucmanio);
+			return process_new(cconfs, p1b, ucmanio);
 		// On Windows, we have to back up the whole file if ctime
 		// changed, otherwise things like permission changes do not get
 		// noticed. So, in that case, fall through to the changed stuff
@@ -321,12 +521,12 @@ static int maybe_do_delta_stuff(struct asfd *asfd,
 	}
 
 	// Got a changed file.
-	//logp("got changed file: %s\n", p1b->path.buf);
+	//logp("got changed file: %s\n", iobuf_to_printable(&p1b->path));
 
 	// If either old or new is encrypted, or librsync is off, we need to
 	// get a new file.
 	// FIX THIS horrible mess.
-	if(!get_int(cconfs[OPT_LIBRSYNC])
+	if(!librsync_enabled(p1b, cb, cconfs)
 	// FIX THIS: make unencrypted metadata use the librsync
 	  || cb->path.cmd==CMD_METADATA
 	  || p1b->path.cmd==CMD_METADATA
@@ -334,7 +534,7 @@ static int maybe_do_delta_stuff(struct asfd *asfd,
 	  || sbuf_is_encrypted(p1b)
 	  || sbuf_is_vssdata(cb)
 	  || sbuf_is_vssdata(p1b))
-		return process_new_file(cconfs, cb, p1b, ucmanio);
+		return process_new(cconfs, p1b, ucmanio);
 
 	// Get new files if they have switched between compression on or off.
 	if(cb->protocol1->datapth.buf
@@ -343,53 +543,59 @@ static int maybe_do_delta_stuff(struct asfd *asfd,
 		oldcompressed=1;
 	if( ( oldcompressed && !compression)
 	 || (!oldcompressed &&  compression))
-		return process_new_file(cconfs, cb, p1b, ucmanio);
+		return process_new(cconfs, p1b, ucmanio);
 
 	// Otherwise, do the delta stuff (if possible).
 	if(sbuf_is_filedata(p1b)
 	  || sbuf_is_vssdata(p1b))
-	{
-		if(process_changed_file(asfd, cconfs, cb, p1b,
-			sdirs->currentdata)) return -1;
-	}
-	else
-	{
-		if(changed_non_file(p1b, ucmanio, p1b->path.cmd, cconfs))
-			return -1;
-	}
-	sbuf_free_content(cb);
-	return 1;
+		return process_changed_file(asfd, cconfs, cb, p1b,
+			sdirs->currentdata);
+	return changed_non_file(p1b, ucmanio, p1b->path.cmd, cconfs);
+}
+
+static enum processed_e process_deleted_file(struct sbuf *cb,
+	struct conf **cconfs)
+{
+	cntr_add_deleted(get_cntr(cconfs), cb->path.cmd);
+	return P_DONE_DELETED;
 }
 
 // return 1 to say that a file was processed
-static int maybe_process_file(struct asfd *asfd,
+static enum processed_e maybe_process_file(struct asfd *asfd,
+	struct dpth *dpth,
 	struct sdirs *sdirs, struct sbuf *cb, struct sbuf *p1b,
-	struct manio *ucmanio, struct conf **cconfs)
+	struct manio *ucmanio, struct manio **hmanio, struct conf **cconfs)
 {
 	int pcmp;
-	if(!(pcmp=sbuf_pathcmp(cb, p1b)))
-		return maybe_do_delta_stuff(asfd, sdirs, cb, p1b,
-			ucmanio, cconfs);
-	else if(pcmp>0)
+	if(p1b)
 	{
-		//logp("ahead: %s\n", p1b->path);
-		// ahead - need to get the whole file
-		if(process_new(cconfs, p1b, ucmanio))
-			return -1;
-		// Do not free.
-		return 1;
+		if(!(pcmp=sbuf_pathcmp(cb, p1b)))
+			return maybe_do_delta_stuff(asfd, dpth,sdirs, cb, p1b,
+				ucmanio, hmanio, cconfs);
+		else if(pcmp>0)
+		{
+			//logp("ahead: %s\n", iobuf_to_printable(&p1b->path));
+			// ahead - need to get the whole file
+			return process_new(cconfs, p1b, ucmanio);
+		}
 	}
-	//logp("behind: %s\n", p1b->path);
+	//logp("behind: %s\n", iobuf_to_printable(&p1b->path));
 	// Behind - need to read more from the old manifest.
 	// Count a deleted file - it was in the old manifest
 	// but not the new.
-	cntr_add_deleted(get_cntr(cconfs), cb->path.cmd);
-	return 0;
+	return process_deleted_file(cb, cconfs);
 }
 
+enum sts_e
+{
+	STS_ERROR=-1,
+	STS_OK=0,
+	STS_BLOCKED=1
+};
+
 // Return 1 if there is still stuff needing to be sent.
 // FIX THIS: lots of repeated code.
-static int do_stuff_to_send(struct asfd *asfd,
+static enum sts_e do_stuff_to_send(struct asfd *asfd,
 	struct sbuf *p1b, char **last_requested)
 {
 	static struct iobuf wbuf;
@@ -399,8 +605,8 @@ static int do_stuff_to_send(struct asfd *asfd,
 		switch(asfd->append_all_to_write_buffer(asfd, &wbuf))
 		{
 			case APPEND_OK: break;
-			case APPEND_BLOCKED: return 1;
-			default: return -1;
+			case APPEND_BLOCKED: return STS_BLOCKED;
+			default: return STS_ERROR;
 		}
 		p1b->flags &= ~SBUF_SEND_DATAPTH;
 	}
@@ -410,8 +616,8 @@ static int do_stuff_to_send(struct asfd *asfd,
 		switch(asfd->append_all_to_write_buffer(asfd, &wbuf))
 		{
 			case APPEND_OK: break;
-			case APPEND_BLOCKED: return 1;
-			default: return -1;
+			case APPEND_BLOCKED: return STS_BLOCKED;
+			default: return STS_ERROR;
 		}
 		p1b->flags &= ~SBUF_SEND_STAT;
 	}
@@ -421,13 +627,13 @@ static int do_stuff_to_send(struct asfd *asfd,
 		switch(asfd->append_all_to_write_buffer(asfd, &wbuf))
 		{
 			case APPEND_OK: break;
-			case APPEND_BLOCKED: return 1;
-			default: return -1;
+			case APPEND_BLOCKED: return STS_BLOCKED;
+			default: return STS_ERROR;
 		}
 		p1b->flags &= ~SBUF_SEND_PATH;
 		free_w(last_requested);
 		if(!(*last_requested=strdup_w(p1b->path.buf, __func__)))
-			return -1;
+			return STS_ERROR;
 	}
 	if(p1b->protocol1->sigjob && !(p1b->flags & SBUF_SEND_ENDOFSIG))
 	{
@@ -443,10 +649,10 @@ static int do_stuff_to_send(struct asfd *asfd,
 			case RS_BLOCKED:
 			case RS_RUNNING:
 				// keep going round the loop.
-				return 1;
+				return STS_BLOCKED;
 			default:
 				logp("error in rs_async: %d\n", sigresult);
-				return -1;
+				return STS_ERROR;
 		}
 	}
 	if(p1b->flags & SBUF_SEND_ENDOFSIG)
@@ -455,12 +661,12 @@ static int do_stuff_to_send(struct asfd *asfd,
 		switch(asfd->append_all_to_write_buffer(asfd, &wbuf))
 		{
 			case APPEND_OK: break;
-			case APPEND_BLOCKED: return 1;
-			default: return -1;
+			case APPEND_BLOCKED: return STS_BLOCKED;
+			default: return STS_ERROR;
 		}
 		p1b->flags &= ~SBUF_SEND_ENDOFSIG;
 	}
-	return 0;
+	return STS_OK;
 }
 
 static int start_to_receive_delta(struct sdirs *sdirs, struct sbuf *rb)
@@ -512,7 +718,8 @@ static int deal_with_receive_end_file(struct asfd *asfd, struct sdirs *sdirs,
 
 	if(fzp_close(&(rb->protocol1->fzp)))
 	{
-		logp("error closing delta for %s in receive\n", rb->path.buf);
+		logp("error closing delta for %s in receive\n",
+			iobuf_to_printable(&rb->path));
 		goto end;
 	}
 	iobuf_move(&rb->endfile, rbuf);
@@ -586,31 +793,28 @@ static int deal_with_filedata(struct asfd *asfd,
 	return 0;
 }
 
+enum str_e
+{
+	STR_ERROR=-1,
+	STR_OK=0,
+	STR_FINISHED=1
+};
+
 // returns 1 for finished ok.
-static int do_stuff_to_receive(struct asfd *asfd,
+static enum str_e do_stuff_to_receive(struct asfd *asfd,
 	struct sdirs *sdirs, struct conf **cconfs,
 	struct sbuf *rb, struct manio *chmanio,
 	struct dpth *dpth, char **last_requested)
 {
 	struct iobuf *rbuf=asfd->rbuf;
 
-	iobuf_free_content(rbuf);
-	// This also attempts to write anything in the write buffer.
-	if(asfd->as->read_write(asfd->as))
-	{
-		logp("error in %s\n", __func__);
-		return -1;
-	}
-
-	if(!rbuf->buf) return 0;
-
 	if(rbuf->cmd==CMD_MESSAGE
 	  || rbuf->cmd==CMD_WARNING)
 	{
 		struct cntr *cntr=NULL;
 		if(cconfs) cntr=get_cntr(cconfs);
 		log_recvd(rbuf, cntr, 0);
-		return 0;
+		return STR_OK;
 	}
 
 	if(rb->protocol1->fzp)
@@ -621,12 +825,12 @@ static int do_stuff_to_receive(struct asfd *asfd,
 			case CMD_APPEND:
 				if(deal_with_receive_append(asfd, rb))
 					goto error;
-				return 0;
+				return STR_OK;
 			case CMD_END_FILE:
 				if(deal_with_receive_end_file(asfd, sdirs, rb,
 					chmanio, cconfs, last_requested))
 						goto error;
-				return 0;
+				return STR_OK;
 			default:
 				iobuf_log_unexpected(rbuf, __func__);
 				goto error;
@@ -638,11 +842,11 @@ static int do_stuff_to_receive(struct asfd *asfd,
 	{
 		case CMD_DATAPTH:
 			iobuf_move(&rb->protocol1->datapth, rbuf);
-			return 0;
+			return STR_OK;
 		case CMD_ATTRIBS:
 			iobuf_move(&rb->attr, rbuf);
 			attribs_decode(rb);
-			return 0;
+			return STR_OK;
 		case CMD_GEN:
 			if(!strcmp(rbuf->buf, "okbackupphase2end"))
 				goto end_phase2;
@@ -656,7 +860,7 @@ static int do_stuff_to_receive(struct asfd *asfd,
 			if(*last_requested
 			  && !strcmp(rbuf->buf, *last_requested))
 				free_w(last_requested);
-			return 0;
+			return STR_OK;
 		default:
 			break;
 	}
@@ -665,14 +869,14 @@ static int do_stuff_to_receive(struct asfd *asfd,
 	{
 		if(deal_with_filedata(asfd, sdirs, rb, rbuf, dpth, cconfs))
 			goto error;
-		return 0;
+		return STR_OK;
 	}
 	iobuf_log_unexpected(rbuf, __func__);
 
 error:
-	return -1;
+	return STR_ERROR;
 end_phase2:
-	return 1;
+	return STR_FINISHED;
 }
 
 static int vss_opts_changed(struct sdirs *sdirs, struct conf **cconfs,
@@ -736,22 +940,138 @@ end:
 	return ret;
 }
 
-// Open the previous (current) manifest.
+// Maybe open the previous (current) manifest.
 // If the split_vss setting changed between the previous backup and the new
 // backup, do not open the previous manifest. This will have the effect of
 // making the client back up everything fresh. Need to do this, otherwise
 // toggling split_vss on and off will result in backups that do not work.
-static int open_previous_manifest(struct manio **cmanio,
+static int maybe_open_previous_manifest(struct manio **manio,
 	struct sdirs *sdirs, const char *incexc, struct conf **cconfs)
 {
-	struct stat statp;
-	if(!lstat(sdirs->cmanifest, &statp)
-	  && !vss_opts_changed(sdirs, cconfs, incexc)
-	  && !(*cmanio=manio_open(sdirs->cmanifest, "rb", PROTO_1)))
+	if(vss_opts_changed(sdirs, cconfs, incexc))
+		return 0;
+	return open_previous_manifest(manio, sdirs);
+}
+
+static int process_next_file_from_manios(struct asfd *asfd,
+	struct dpth *dpth,
+	struct sdirs *sdirs,
+	struct manio **p1manio,
+	struct manio **cmanio,
+	struct manio *ucmanio,
+	struct manio **hmanio,
+	struct sbuf **p1b,
+	struct sbuf *cb,
+	struct conf **cconfs)
+{
+	if(!(*p1b=sbuf_alloc(PROTO_1)))
+		goto error;
+	switch(manio_read(*p1manio, *p1b))
 	{
-		logp("could not open old manifest %s\n", sdirs->cmanifest);
-		return -1;
+		case 0:
+			break;
+		case 1:
+			manio_close(p1manio);
+			if(asfd->write_str(asfd, CMD_GEN, "backupphase2end"))
+				goto error;
+			sbuf_free(p1b);
+			break;
+		case -1:
+			goto error;
 	}
+
+	if(!*cmanio)
+	{
+		if(!*p1b) return 0;
+		// No old manifest, need to ask for a new file.
+		switch(process_new(cconfs, *p1b, ucmanio))
+		{
+			case P_NEW:
+				return 0;
+			case P_DONE_NEW:
+				goto p1b_done;
+			default:
+				goto error;
+		}
+	}
+
+	// Have an old manifest, look for it there.
+
+	// Might already have it, or be ahead in the old
+	// manifest.
+	if(cb->path.buf)
+	{
+		switch(maybe_process_file(asfd, dpth,
+			sdirs, cb, *p1b, ucmanio, hmanio, cconfs))
+		{
+			case P_NEW:
+				return 0;
+			case P_CHANGED:
+				// Free cb content for things like encrypted
+				// files, which we only pretend are new.
+				sbuf_free_content(cb);
+				return 0;
+			case P_DONE_NEW:
+				goto p1b_done;
+			case P_DONE_CHANGED:
+			case P_DONE_UNCHANGED:
+				sbuf_free_content(cb);
+				goto p1b_done;
+			case P_DONE_DELETED:
+				sbuf_free_content(cb);
+				break;
+			case P_ERROR:
+				goto error;
+		}
+	}
+
+	while(1)
+	{
+		sbuf_free_content(cb);
+		switch(manio_read(*cmanio, cb))
+		{
+			case 0: break;
+			case 1: manio_close(cmanio);
+				if(*p1b) switch(process_new(cconfs,
+					*p1b, ucmanio))
+				{
+					case P_NEW:
+						return 0;
+					case P_DONE_NEW:
+						goto p1b_done;
+					default:
+						goto error;
+				}
+				return 0;
+			case -1: goto error;
+		}
+		switch(maybe_process_file(asfd, dpth, sdirs,
+			cb, *p1b, ucmanio, hmanio, cconfs))
+		{
+			case P_NEW:
+				return 0;
+			case P_CHANGED:
+				// Free cb content for things like encrypted
+				// files, which we only pretend are new.
+				sbuf_free_content(cb);
+				return 0;
+			case P_DONE_NEW:
+				goto p1b_done;
+			case P_DONE_CHANGED:
+			case P_DONE_UNCHANGED:
+				sbuf_free_content(cb);
+				goto p1b_done;
+			case P_DONE_DELETED:
+				sbuf_free_content(cb);
+				continue;
+			case P_ERROR:
+				goto error;
+		}
+	}
+error:
+	return -1;
+p1b_done:
+	sbuf_free(p1b);
 	return 0;
 }
 
@@ -766,7 +1086,8 @@ int backup_phase2_server_protocol1(struct async *as, struct sdirs *sdirs,
 	char *last_requested=NULL;
 	struct manio *chmanio=NULL; // changed data
 	struct manio *ucmanio=NULL; // unchanged data
-	struct manio *cmanio=NULL; // previous (current) manifest.
+	struct manio *cmanio=NULL; // previous (current) manifest
+	struct manio *hmanio=NULL; // to look up deleted hardlinks
 	struct sbuf *cb=NULL; // file list in current manifest
 	struct sbuf *p1b=NULL; // file list from client
 	struct sbuf *rb=NULL; // receiving file from client
@@ -812,7 +1133,13 @@ int backup_phase2_server_protocol1(struct async *as, struct sdirs *sdirs,
 		get_int(cconfs[OPT_MAX_STORAGE_SUBDIRS])))
 			goto error;
 
-	if(open_previous_manifest(&cmanio, sdirs, incexc, cconfs))
+	if(maybe_open_previous_manifest(&cmanio, sdirs, incexc, cconfs))
+		goto error;
+	// If the first file that others hardlink to has been deleted, we will
+	// need to look through the previous manifest to find the information
+	// for that original file, in order to not have to copy all the data
+	// across again.
+	if(open_previous_manifest(&hmanio, sdirs))
 		goto error;
 
 	if(get_int(cconfs[OPT_DIRECTORY_TREE]))
@@ -830,7 +1157,7 @@ int backup_phase2_server_protocol1(struct async *as, struct sdirs *sdirs,
 	{
 		if(!(p1pos=do_resume(sdirs, dpth, cconfs)))
 			goto error;
-		if(cntr_send_sdirs(asfd, sdirs, cconfs))
+		if(cntr_send_sdirs(asfd, sdirs, cconfs, CNTR_STATUS_BACKUP))
 			goto error;
 	}
 
@@ -856,83 +1183,47 @@ int backup_phase2_server_protocol1(struct async *as, struct sdirs *sdirs,
 			return breakpoint(breaking, __func__);
 
 		if(write_status(CNTR_STATUS_BACKUP,
-			rb->path.buf?rb->path.buf:p1b->path.buf, cntr))
+			rb->path.buf?rb->path.buf:"", cntr))
 				goto error;
 		if(last_requested
 		  || !p1manio
 		  || asfd->writebuflen)
 		{
-			switch(do_stuff_to_receive(asfd, sdirs,
+			iobuf_free_content(asfd->rbuf);
+			if(asfd->as->read_write(asfd->as))
+			{
+				logp("error in %s\n", __func__);
+				goto error;
+			}
+
+			if(asfd->rbuf->buf)
+			  switch(do_stuff_to_receive(asfd, sdirs,
 				cconfs, rb, chmanio, dpth, &last_requested))
 			{
-				case 0: break;
-				case 1: goto end; // Finished ok.
-				case -1: goto error;
+				case STR_OK:
+					break;
+				case STR_FINISHED:
+					goto end;
+				case STR_ERROR:
+					goto error;
 			}
 		}
 
-		switch(do_stuff_to_send(asfd, p1b, &last_requested))
+		if(p1b) switch(do_stuff_to_send(asfd, p1b, &last_requested))
 		{
-			case 0: break;
-			case 1: continue;
-			case -1: goto error;
-		}
-
-		if(!p1manio) continue;
-
-		sbuf_free_content(p1b);
-
-		switch(manio_read(p1manio, p1b))
-		{
-			case 0: break;
-			case 1: manio_close(&p1manio);
-				if(asfd->write_str(asfd,
-				  CMD_GEN, "backupphase2end")) goto error;
+			case STS_OK:
+				sbuf_free(&p1b);
 				break;
-			case -1: goto error;
-		}
-
-		if(!cmanio)
-		{
-			// No old manifest, need to ask for a new file.
-			if(process_new(cconfs, p1b, ucmanio))
+			case STS_BLOCKED:
+				continue;
+			case STS_ERROR:
 				goto error;
-			continue;
 		}
 
-		// Have an old manifest, look for it there.
-
-		// Might already have it, or be ahead in the old
-		// manifest.
-		if(cb->path.buf) switch(maybe_process_file(asfd,
-			sdirs, cb, p1b, ucmanio, cconfs))
-		{
-			case 0: break;
-			case 1: continue;
-			case -1: goto error;
-		}
-
-		while(cmanio)
-		{
-			sbuf_free_content(cb);
-			switch(manio_read(cmanio, cb))
-			{
-				case 0: break;
-				case 1: manio_close(&cmanio);
-					if(process_new(cconfs, p1b,
-						ucmanio)) goto error;
-					continue;
-				case -1: goto error;
-			}
-			switch(maybe_process_file(asfd, sdirs,
-				cb, p1b, ucmanio, cconfs))
-			{
-				case 0: continue;
-				case 1: break;
-				case -1: goto error;
-			}
-			break;
-		}
+		if(p1manio
+		 && process_next_file_from_manios(asfd, dpth, sdirs,
+			&p1manio, &cmanio, ucmanio, &hmanio, &p1b, cb, cconfs))
+					goto error;
 	}
 
 error:
@@ -955,6 +1246,7 @@ end:
 	sbuf_free(&rb);
 	manio_close(&p1manio);
 	manio_close(&cmanio);
+	manio_close(&hmanio);
 	dpth_free(&dpth);
 	man_off_t_free(&p1pos);
 	if(!ret && sdirs)
